#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy
import random
import sys
import time
from deap import tools
from utils import *



class GeneticAlgorithm(object):


  def __init__(self, matrix, type_selection="tournament", \
               type_crossing="2points", \
               type_mutation="reciprocal_interchange", \
               type_reemplace="elitistic", \
               type_condition="max_generations", stop_elapsed_time=None, \
               elitism=None, fitness_goal=None, known_optim=None, \
               initial_population=None, max_population=50, \
               repeated_chromosomes=True, function_criteria="MIN"):
    # repeated = duplicated
    self.matrix = matrix
    self.function_criteria = function_criteria

    if initial_population == None:
      if max_population == None:
        print("ERROR, NOT GIVEN initial_population OR max_population")
        sys.exit(0)
      # minimum two single elements in population, maximum max_population
      num_initial_population = random.randint(2, max_population)
      self.parents = get_n_permutations(len(matrix), \
                                        number=num_initial_population, \
                                        repeated=repeated_chromosomes)
      print("PARENTS: ")
      for e in self.parents:
        print(e)
      print()
    else:
      self.parents = initial_population

    self.childs = []
    self.repeated_chromosomes = repeated_chromosomes

    self.num_generation = 0
    self.best_solution = best_chromosome(self.matrix, self.parents, \
                                         self.function_criteria)
    self.iteration_best_solution = 0
    self.start_time = time.time()
    self.stop_elapsed_time = stop_elapsed_time

    self.type_selection = type_selection
    self.type_crossing = type_crossing
    self.type_mutation = type_mutation
    self.type_reemplace = type_reemplace
    self.type_condition = type_condition
    self.elitism = elitism
    if self.elitism != None:
      if self.elitism > len(self.parents):
        print("ERROR self.elitism > len(self.parents)")
        #self.elitism = len(self.parents)
        sys.exit(0)
      else:
        self.elitist_set = {}
    self.fitness_goal = fitness_goal
    self.known_optim = known_optim



  def execute(self):
    """
    Execute genetic algorithm.
    """
    while self.stop_condition():
      print("\n\n")
      print("\x1b[6;30;42mGENERATION {}\x1b[0m" \
            "\n".format(self.num_generation))
      print(self.cross())
      print(self.mutation())
      print()
      self.reemplace_population()
      #print("\n\n")



  def selection(self, num_pairs=None):
    """
    Make populations's selection. 

    @param num_pairs    number of pairs to cross with parents
    @return pairs       list of Chromosome pairs to cross it after
    """
    if num_pairs == None:
      num_pairs = len(self.parents)
    r = roulette(self.matrix, self.parents, self.function_criteria)

    if self.type_selection == "estocastical":
      # roulette with several pointers uniform distributed
      num_pointers = num_pairs * 2
      aux = 360 / num_pointers
      pointers_list = [0]

      i = 0
      while i < num_pointers - 1:
        pointers_list.append(pointers_list[i] + aux)
        i = i + 1
      
      i = 0
      aux = random.randint(0, 360)
      while i < num_pointers:
        pointers_list[i] = (pointers_list[i] - aux) % 360
        i = i + 1

      pointers_list.sort()

      selected_parents = []
      for x in pointers_list:
        for chromosome in r:
          if r[chromosome][0] <= x <= r[chromosome][1]:
            selected_parents.append(chromosome)
            break

      pairs = []
      i = 0
      while i < num_pairs:
        parent_1 = random.choice(selected_parents)
        parent_2 = random.choice(selected_parents)
        # to make pairs without repeated chromsomes in pair
        """
        while parent_1 == parent_2:
          parent_2 = random.choice(selected_parents)
        """
        pairs.append([list(parent_1), list(parent_2)])
        i = i + 1

    elif self.type_selection == "tournament":
      # roulette with only one pointer
      pair = []
      pairs = []
      for x in range(num_pairs * 2):
        random_number = random.random() * 360
        for chromosome in r:
          if r[chromosome][0] < random_number < r[chromosome][1]:
            if len(pair) == 1:
              pair.append(list(chromosome))
              pairs.append(pair)
              pair = []
            else:
              pair.append(list(chromosome))
            break

    else:
      print("ERROR IN type_selection")
      sys.exit(0)

    print("SELECTION: len -> {}".format(len(pairs)))
    return pairs



  def cross(self, prob=1):
    """
    Cross selection (pair single element) to make childs.
    """
    if self.type_crossing == "2points":
      selected_parents = self.selection()
      for pair in selected_parents:
        cut_point_1 = random.randint(0, len(pair[0]))
        cut_point_2 = random.randint(0, len(pair[0]))
        while cut_point_1 == cut_point_2:   # point_1 has to be != point_2
          cut_point_2 = random.randint(0, len(pair[0]))
        if cut_point_1 > cut_point_2:   # ordering points
          aux = cut_point_1
          cut_point_1 = cut_point_2
          cut_point_2 = aux

        child = list(pair[0][:cut_point_1] + pair[0][cut_point_2:])
        print("[{} - {}] - ({}, {})".format(pair[0], pair[1], cut_point_1, \
                                            cut_point_2))
        print("Intermediate child: {}".format(child))

        missing_cities = set(pair[0])
        missing_cities = list(missing_cities.difference(set(child)))
        print("Missing cities: {}".format(missing_cities))

        middle_part = []
        # check what chromosome is better in the pair and take it to take 
        # missing cities order in this chromosome
        if fitness_one(self.matrix, pair[0], self.function_criteria) < \
           fitness_one(self.matrix, pair[1], self.function_criteria):
          for x in pair[0]:
            if x in missing_cities:
              middle_part.append(x)
        else:
          for x in pair[1]:
            if x in missing_cities:
              middle_part.append(x)

        child = list(pair[0][:cut_point_1]) + middle_part + \
                list(pair[0][cut_point_2:])
        print(child)
        self.childs.append(child)


    elif self.type_crossing == "1point":
      pass


    elif self.type_crossing == "uniform":
      selected_parents = self.selection()
      for pair in selected_parents:
        pass


    elif self.type_crossing == "heuristic":
      pass
    

    elif self.type_crossing == "partially_mapped":
      selected_parents = self.selection(int(len(self.parents) / 2))
      for pair in selected_parents:
        aux = tools.cxPartialyMatched(pair[0], pair[1])
        self.childs.append(aux[0])
        self.childs.append(aux[1])


    elif self.type_crossing == "order":
      pass
    

    elif self.type_crossing == "cycle":
      pass


    else:
      print("ERROR IN type_crossing")
      sys.exit(0)


    print("\nCROSS: len -> {}".format(len(self.childs)))
    return self.childs



  def mutation(self, prob=4.9):
    """
    prob => 0.1 and prob <= 5 by default.
    """
    if self.type_mutation == "insertion":
      for child in self.childs:
        prob_mask = [random.random() * 100 for x in range(len(child))]
        i = 0
        while i < len(prob_mask):
          if prob_mask[i] < prob:   # mutation
            prob_mask[i] = True
          else:
            prob_mask[i] = False
          i = i + 1

        i = 0
        while i < len(prob_mask):
          if prob_mask[i]:
            child.insert(random.randint(0, len(child)), child.pop(i))
          i = i + 1


    elif self.type_mutation == "reciprocal_interchange":
      for child in self.childs:
        pos = random.randint(0, len(child) - 1)
        pos_1 = random.randint(0, len(child) - 1)
        while pos == pos_1:   # pos has to be != pos_1
          pos_1 = random.randint(0, len(child) - 1)
        # interchange gens
        temp = child[pos]
        child[pos] = child[pos_1]
        child[pos_1] = temp


    else:
      print("ERROR IN type_mutation")
      sys.exit(0)


    # we have to remove possible chromosomes repeated if we do not allow
    # repeated chromosomes in poblation
    if self.repeated_chromosomes == False:
      aux_set = set(tuple(chromosome) for chromosome in self.childs)
      self.childs = [list(chromosome) for chromosome in aux_set]
    print("\nMUTATION: len -> {}".format(len(self.childs)))
    return self.childs



  def reemplace_population(self):
    if self.type_reemplace == "generational":
      # remove chromosomes with infinite fitness
      for chromosome in self.childs:
        if fitness_one(self.matrix, chromosome, \
                       self.function_criteria) == numpy.inf or \
           fitness_one(self.matrix, chromosome,   
                       self.function_criteria) == -numpy.inf:
          while chromosome in self.childs:
            self.childs.remove(chromosome)

      # if len(self.childs) < len(self.parents), we generate valid 
      # chromosomes to avoid that poblation decreases
      while len(self.childs) < len(self.parents):
        random_chromosome = random.sample(list(range(0, len(self.matrix))), \
                                          len(self.matrix))
        if self.repeated_chromosomes:
          while fitness_one(self.matrix, random_chromosome,   
                            self.function_criteria) == numpy.inf or \
                fitness_one(self.matrix, random_chromosome, \
                            self.function_criteria) == -numpy.inf:
            random_chromosome = random.sample(list(range(0, \
                                                   len(self.matrix))), \
                                              len(self.matrix))
        else:
          while random_chromosome in self.childs or \
                fitness_one(self.matrix, random_chromosome, \
                            self.function_criteria) == numpy.inf or \
                fitness_one(self.matrix, random_chromosome, \
                            self.function_criteria) == -numpy.inf:
            random_chromosome = random.sample(list(range(0, \
                                                   len(self.matrix))), \
                                              len(self.matrix))
        self.childs.append(random_chromosome)
      
      """
      aux = []
      for chromosome in self.childs:
        aux.append((tuple(chromosome), fitness_one(self.matrix, chromosome, \
                                                   self.function_criteria)))
      if len(self.elitist_set) != 0:   # after first passing
        aux = aux + list(self.elitist_set.items())
        self.elitist_set = {}
      aux = sorted(aux, key=lambda x : x[1], reverse=False)
      i = 0
      while i < self.elitism:
        self.elitist_set[aux[i][0]] = aux[i][1]
        i += 1
      print(self.elitist_set)
      """

      self.parents = self.childs
      self.childs = []


    elif self.type_reemplace == "steady_state_ag":
      pass


    elif self.type_reemplace == "enpacked":
      pass
    

    elif self.type_reemplace == "final_judgement":
      pass


    elif self.type_reemplace == "catastrofic_reemplace":
      pass


    elif self.type_reemplace == "elitistic":
      aux = []
      for chromosome in self.childs:
        aux.append((chromosome, fitness_one(self.matrix, chromosome, \
                                            self.function_criteria)))
      if len(self.elitist_set) != 0:   # after first passing
        aux = aux + random.shuffle(self.elitist_set.items())
        self.elitist_set = {}
      aux = sorted(aux, key=lambda x : x[1], reverse=False)
      i = 0
      while i < self.elitism:
        self.elitist_set[tuple(aux[i][0])] = aux[i][1]
        i += 1
      #print(self.elitist_set)
      aux_1 = random.randint(0, self.elitism)
      aux = aux[self.elitism - aux_1:-aux_1]
      self.parents = []
      for element in aux:
        self.parents.append(list(element[0]))
      self.childs = []
      #print(self.parents)


    else:
      print("ERROR IN type_reemplace")
      sys.exit(0)



  def stop_condition(self):
    # choose best chromosome in this iteration and compare it with best solution
    # known in all execution
    aux = best_chromosome(self.matrix, self.parents, self.function_criteria)
    print("BEST CHROMOSOME IN THIS ROUND: {} - " \
          "{}".format(aux, fitness_one(self.matrix, aux, \
                      self.function_criteria)))
    print()
    print("BEST SOLUTION ACTUALLY: {} - " \
          "{}".format(self.best_solution, \
                      fitness_one(self.matrix, self.best_solution, \
                                  self.function_criteria)))
    if fitness_one(self.matrix, aux, self.function_criteria) < \
       fitness_one(self.matrix, self.best_solution, \
                   self.function_criteria) and \
       fitness_one(self.matrix, aux, \
                   self.function_criteria) != numpy.inf and \
       fitness_one(self.matrix, aux, self.function_criteria) != -numpy.inf:
      self.best_solution = aux
      self.iteration_best_solution = self.num_generation


    if self.type_condition == "max_generations":   # FINISHED
      if self.num_generation == 600:
        return False
      else:
        self.num_generation = self.num_generation + 1
        #input()
        return True


    elif self.type_condition == "execution_time":   # FINISHED
      # type(elapsed_time) -> float(seconds)
      elapsed_time = time.time() - self.start_time
      if elapsed_time > self.stop_elapsed_time:
        sys.exit(0)


    elif self.type_condition == "convergency":
      pass


    elif self.type_condition == "known_optim":   # FINISHED
      if fitness_one(self.matrix, self.best_solution, \
                     self.function_criteria) == self.known_optim:
        sys.exit(0)


    elif self.type_condition == "fitness":   # FINISHED
      if fitness_one(self.matrix, self.best_solution, \
                     self.function_criteria) < self.fitness_goal:
        sys.exit(0)


    else:
      print("ERROR IN type_condition")
      sys.exit(0)




if __name__ == "__main__":
  m = numpy.array([[-numpy.inf, 3, 4, 2, 6, 1, -numpy.inf, 7, 14, 8, 5, 3], \
                   [2, -numpy.inf, 6, 13, 3, 4, 6, 1, 3, 12, 6, 4], \
                   [1, 9, -numpy.inf, 3, 3, 6, 12, 6, 10, 3, 3, 12], \
                   [1, 3, 5, -numpy.inf, 5, 8, 8, 4, 2, 9, 4, 9], \
                   [3, 5, 2, 5, -numpy.inf, 7, 4, -numpy.inf, 1, 12, 7, 13], 
                   [7, 4, 6, 15, 3, -numpy.inf, 10, 2, 3, 4, 6, 5], \
                   [4, 1, 7, 11, 4, 5, -numpy.inf, 11, 10, 6, 4, 1], \
                   [1, 3, 3, 6, 2, 1, 6, -numpy.inf, 9, 1, 8, 6], \
                   [4, 6, 4, 5, 7, 6, 14, 6, -numpy.inf, 9, 1, 1], \
                   [3, 10, 2, 8, 4, 9, 1, 1, 4, -numpy.inf, 1, 9], \
                   [1, 4, 5, 9, 1, 6, 1, 5, 15, 4, -numpy.inf, 5], \
                   [11, 3, -numpy.inf, 7, 3, 2, 5, 9, 11, 8, 5, -numpy.inf]])
  population = [[2, 4, 0, 3, 1, 9, 11, 7, 6, 8, 5, 10], \
                [2, 4, 0, 3, 1, 9, 11, 7, 6, 8, 5, 10], \
                [1, 8, 2, 6, 9, 5, 10, 0, 3, 11, 4, 7], \
                [10, 0, 11, 5, 4, 7, 6, 9, 1, 8, 3, 2], \
                [10, 0, 11, 5, 4, 7, 6, 9, 1, 8, 3, 2], \
                [3, 8, 5, 2, 11, 0, 1, 9, 7, 6, 4, 10]]

  # ONLY USE CHROMOSOMES AS TUPLE WHERE WE USE DICT OR SET!!!
  g = GeneticAlgorithm(matrix=m, initial_population=population, \
                       type_selection="tournament", \
                       type_reemplace="generational", \
                       repeated_chromosomes=True, function_criteria="MAX")
  """
  g = GeneticAlgorithm(matrix=m, type_selection="tournament", \
                       type_reemplace="elitistic", elitism=3, \
                       repeated_chromosomes=True, function_criteria="MAX")
  """
  g.execute()
  print("\n\n\n\nBEST CHROMOSOME: {}\nFITNESS: {}\nN_GENERATION: " \
        "{}".format(g.best_solution, fitness_one(m, g.best_solution), \
                                                 g.iteration_best_solution))
