#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import math
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
import numpy
import random
from collections import Counter
from deap import tools
from fractions import Fraction
from functools import reduce
from matplotlib.patches import Rectangle
from mpl_toolkits.mplot3d import Axes3D
from operator import mul



def _map(swapped, cross_points):
  """
  Subroutine for pmx() function
  """
  n = len(swapped[0])
  c1, c2 = cross_points
  print(cross_points)
  s1, s2 = swapped
  map_ = s1[c1:c2], s2[c1:c2]
  print(map_)
  for i_chromosome in range(n):
    if not c1 < i_chromosome < c2:
      for i_son in range(2):
        while _repeated(swapped[i_son][i_chromosome], swapped[i_son]):
          map_index = map_[i_son].index(swapped[i_son][i_chromosome])
          swapped[i_son][i_chromosome] = map_[1-i_son][map_index]

  return s1, s2



def _repeated(element, collection):
  """
  Subroutine for pmx() function
  """
  c = 0
  for e in collection:
    if e == element:
      c += 1

  return c > 1



def _swap(data_a, data_b, cross_points):
  """
  Subroutine for pmx() function
  """
  c1, c2 = cross_points
  new_a = data_a[:c1] + data_b[c1:c2] + data_a[c2:]
  new_b = data_b[:c1] + data_a[c1:c2] + data_b[c2:]

  return new_a, new_b



def best_chromosome(matrix, population, function_criteria="MIN"):
  """
  Returns chromosome with best fitness from population.

  @param matrix               distance matrix
  @param population           list of chromosomes
  @param function_criteria    minimize/maximize fitness function
  @return res                 best chromosome  
  """
  res = population[0]
  for chromosome in population[1:]:
    if fitness_one(matrix, chromosome, function_criteria) < \
       fitness_one(matrix, res, function_criteria):
      res = chromosome

  return res



def crowding_distance(front):
  """
  Calculates a crowding distance maked in NSGAII

  @param front         list with chromosomes that makes Pareto's front
  @return distances    a dict with crowding distance for each chromosome
  """
  distances = {}
  for chromosome in front:
    distances[chromosome] = 0
  i = 0
  while i < len(list(front)[0]):   # for each objective
    q = sorted(list(front), key=lambda x : x[i])
    distances[q[0]] = math.inf
    distances[q[len(q) - 1]] = math.inf
    f_max = max([chromosome[i] for chromosome in q])
    f_min = min([chromosome[i] for chromosome in q])
    j = 1
    while j < len(q) - 1:
      distances[q[j]] += (q[j + 1][i] - q[j - 1][i]) / (f_max - f_min)
      j += 1
    i += 1

  return distances



def dominate(chromosome_1, chromosome_2, criteria_function="MIN"):
  """
  Returns 1 if chromosome_1 domains chromosome_2
  Returns -1 if chromosome_2 domains chromosome_1
  Returns 0 if nobody domains other
  """
  aux = [None] * len(chromosome_1)
  i = 0
  while i < len(chromosome_1):
    if chromosome_1[i] > chromosome_2[i]:
      if criteria_function == "MAX":
        aux[i] = True
      elif criteria_function == "MIN":
        aux[i] = False
    elif chromosome_1[i] < chromosome_2[i]:
      if criteria_function == "MAX":
        aux[i] = False
      elif criteria_function == "MIN":
        aux[i] = True
    i = i + 1
  
  if False in aux:
    if True in aux:
      return 0
    else:
      return -1
  elif True in aux:
    return 1
  else:
    return 0



def factorial(number):
  """
  Calculate a factorial from a number.

  @param number    number to calculate factorial
  @return          factorial's number
  """
  return reduce(mul, list(range(1, number + 1)))



def fitness_all(matrix, population, function_criteria="MIN"):
  """
  Calculate fitness from all chromosomes from population.

  @param matrix               distance matrix
  @param population           list of chromosomes
  @param function_criteria    minimize/maximize fitness function
  @return res                 dict with all chromosomes with his fitness
  """
  res = {}
  for chromosome in population:
    res[tuple(chromosome)] = 0
    i = 0
    while i < len(chromosome) - 1:
      res[tuple(chromosome)] += matrix[chromosome[i]][chromosome[i + 1]]
      i += 1
    if res[tuple(chromosome)] != numpy.inf and \
       res[tuple(chromosome)] != -numpy.inf:
      res[tuple(chromosome)] = int(res[tuple(chromosome)])
    if function_criteria == "MAX":
      res[tuple(chromosome)] = -res[tuple(chromosome)]

  print("FITNESS: ")
  for e in res:
    print("{} -> {}".format(list(e), res[e]))
  print()
  
  return res



def fitness_one(matrix, chromosome, function_criteria="MIN"):
  """
  Calculate fitness from one chromosome.

  @param matrix               distance matrix
  @param chromosome           chromosome that want to calculate fitness
  @param function_criteria    minimize/maximize fitness function
  @return res                 fitness from chromosome
  """
  # WE CAN CHANGE FITNESS SIGN HERE DESPITE OF MATRIX
  res = 0
  i = 0
  while i < len(chromosome) - 1:
    res += matrix[chromosome[i]][chromosome[i + 1]]
    i += 1
  if res != numpy.inf and res != -numpy.inf:
    res = int(res)

  if function_criteria == "MAX":
    res = -res

  return res



def get_n_permutations(length, number, repeated=False):
  """
  Returns a list of number lists with length permutations.

  @param length      length from every permutation
  @param number      number of permutations that want
  @param repeated    let return repeated permutations
  @return res        list of permutations
  """
  aux = list(range(0, length))
  res = []
  if repeated:
    for _ in range(number):
      res.append(random.sample(aux, len(aux)))
  else:
    fac = factorial(len(aux))
    if fac < number:
      raise Exception("You cannot generate more permutations than " \
                      "{}".format(fac))
    else:
      while len(res) < number:
        s = random.sample(aux, len(aux))
        while s in res:
          s = random.sample(aux, len(aux))
        res.append(s)

  res.sort()
  return res



def non_dominated_sort(population):
  """
  Method that sort chromosomes in NSGAII

  @param population    a list of chromosomes
  @return res          a ordered list with chromosome's set that represents
                       a Pareto Front excluding his left sets in res
  """
  res = [None, set()]
  dict_aux = {}
  for p in population:
    dict_aux[p] = [set(), 0]   # Sp, np and rank
    for q in population:
      if dominate(p, q) == 1:
        dict_aux[p][0].add(q)
      elif dominate(p, q) == -1:
        dict_aux[p][1] += 1
    if dict_aux[p][1] == 0:
      res[1].add(p)
      dict_aux[p].append(1)
  i = 1
  while len(res[i]) != 0:
    Q = set()
    for p in res[i]:
      for q in dict_aux[p][0]:
        dict_aux[q][1] -= 1
        if dict_aux[q][1] == 0:
          dict_aux[q].append(i + 1)
          Q.add(q)
    i = i + 1
    res.append(Q)
  
  return res[1:-1]



def pareto_optimal(points, function_domains=None):
  if function_domains == None:
    function_domains = ["MAX"] * len(points[0])
  DIC_POINTS = {}   # auxiliar variable, not used
  for i in range(0, len(points)):
    DIC_POINTS[i] = points[i]
  #print(DIC_POINTS)

  mat = numpy.array([[None] * len(points)] * len(points))

  i = 0
  while i < len(points):
    j = i + 1
    while j < len(points):
      aux = [None] * len(points[i])
      k = 0
      while k < len(points[i]):
        if function_domains[k] == "MAX":
          if points[i][k] > points[j][k]:
            aux[k] = True
          elif points[i][k] < points[j][k]:
            aux[k] = False
        elif function_domains[k] == "MIN":
          if points[i][k] > points[j][k]:
            aux[k] = False
          elif points[i][k] < points[j][k]:
            aux[k] = True
        k = k + 1
      if False in aux:
        if True in aux:
          aux = 0
        else:
          aux = -1
      elif True in aux:
        aux = 1
      else:
        aux =  0
      mat[i][j] = aux
      j = j + 1
    i = i + 1

  for i in range(0, len(points)):
    mat[i][i] = 0

  for i in range(0, len(points)):
    for j in range(i, len(points)):
      mat[j][i] = -mat[i][j]

  print(mat)   # adjacent matrix

  pareto_points = []
  i = 0
  while i < len(points):
    if -1 not in mat[i]:
      pareto_points.append(points[i])
      #print(mat[i])
    i = i + 1
  #print(pareto_points)

  # divide pareto points and other points
  others_points = []
  for x in points:
    found = False
    for y in pareto_points:
      if numpy.array_equal(x, y):
        found = True
        continue
    if not found:
      others_points.append(x)

  return pareto_points, others_points



def plot_pareto(pareto_points, others_points, function_domains=None):
  #print(others_points)
  #print(others_points[0][0])
  if function_domains == None:
    function_domains = ["MAX"] * len(pareto_points[0])
  
  # plot graphics for 2D and 3D
  if len(function_domains) == 2:
    plt.figure()
    ax = plt.subplot(111)
    ax.set_xlabel("Dimension x")
    ax.set_ylabel("Dimension y")
    ax.set_xlim([0, 11])
    ax.set_ylim([0, 11])
    ax.xaxis.set_ticks(range(0, 11, 1))
    ax.yaxis.set_ticks(range(0, 11, 1))
    box = ax.get_position()
    ax.set_position([box.x0, box.y0 + box.height * 0.1, box.width, \
                     box.height * 0.9])

    ax.plot([x[0] for x in others_points], [y[1] for y in others_points], \
            "bo", label="others points")

    # sort pareto points to print frontier correctly
    if function_domains[0] == "MAX":
      if function_domains[1] == "MAX":
        pareto_points.sort(key=lambda x: (x[0], x[1]), reverse=True)
      elif function_domains[1] == "MIN":
        pareto_points.sort(key=lambda x: (x[0], -x[1]), reverse=True)
    elif function_domains[0] == "MIN":
      if function_domains[1] == "MAX":
        pareto_points.sort(key=lambda x: (x[0], -x[1]))
      elif function_domains[1] == "MIN":
        pareto_points.sort(key=lambda x: (x[0], x[1]))
    if len(pareto_points) > 1:
      ax.plot([x[0] for x in pareto_points], \
              [y[1] for y in pareto_points], "g-", \
              label="frontier pareto (aprox)")

      ax.plot([x[0] for x in pareto_points], \
              [y[1] for y in pareto_points], "rx", label="optimal pareto")

    plt.legend(loc="upper center", bbox_to_anchor=(0.5, 1.17))
    ax.grid(color=(0.9, 0.9, 0.9), linestyle="-", linewidth=0.5)
    plt.show()
  elif len(function_domains) == 3:
    plt.figure()
    ax = plt.subplot(111, projection="3d")
    ax.set_xlabel("Dimension x")
    ax.set_ylabel("Dimension y")
    ax.set_zlabel("Dimension z")
    ax.set_xlim3d([0, 11])
    ax.set_ylim3d([0, 11])
    ax.set_zlim3d([0, 11])
    ax.xaxis.set_ticks(range(0, 11, 1))
    ax.yaxis.set_ticks(range(0, 11, 1))
    ax.zaxis.set_ticks(range(0, 11, 1))
    box = ax.get_position()
    ax.set_position([box.x0, box.y0 + box.height * 0.1, box.width, \
                     box.height * 0.9])
        
    ax.scatter([x[0] for x in others_points], \
               [y[1] for y in others_points], \
               [z[2] for z in others_points], color="blue", \
               label="other points", marker="o")
    
    ax.scatter([x[0] for x in pareto_points], \
               [y[1] for y in pareto_points], \
               [z[2] for z in pareto_points], color="red", \
               label="optimal pareto", marker="x")
    
    if len(list(pareto_points)) == 2:
      ax.plot3D([x[0] for x in pareto_points], \
                [y[1] for y in pareto_points], \
                [z[2] for z in pareto_points], "g-")
    elif len(list(pareto_points)) >= 3:
      pp = numpy.array(list(pareto_points))
      triang = mtri.Triangulation(pp[:,0], pp[:,1])
      #ax.plot_trisurf(triang, pp[:,2], alpha=0.35, color="green", \
      #                edgecolor="black")
      ax.plot_trisurf(triang, pp[:,2], alpha=0.35, color="green")
      

    p1 = Rectangle((0, 0), 1, 1, fc="blue")
    p2 = Rectangle((0, 0), 1, 1, fc="red")
    if len(list(pareto_points)) != 1:
      p3 = Rectangle((0, 0), 1, 1, fc="green")
      ax.legend([p1, p2, p3], ["other points", "pareto optimal", \
                               "pareto frontier (aprox)"], \
                loc="center right", bbox_to_anchor=(1.1, 0.5))
    elif len(list(pareto_points)) == 1:
      ax.legend([p1, p2], ["other points", "pareto optimal"], \
                loc="center right", bbox_to_anchor=(1.1, 0.5))
    ax.grid(color=(0.9, 0.9, 0.9), linestyle="-", linewidth=0.5)
    ax.view_init(35, -125)
    plt.show()



def plot_jMetal_data():
  """
  Method that plots data from jMetal algorithm
  """
  fitness = [(float(line.rstrip("\n").split(" ")[0]), \
              float(line.rstrip("\n").split(" ")[1])) \
             for line in open("FUN.tsv", "r")]
  chromosomes = [tuple(line.rstrip("\n").split()) \
                 for line in open("VAR.tsv", "r")]
  i = 0
  res = []
  while i < len(fitness):
    res.append((chromosomes[i], fitness[i]))
    i += 1
  print(res)
  f = [e[1] for e in res]
  l = sorted(f, key=lambda x : (x[0], x[1]))
  plt.plot([e[0] for e in l],
           [e[1] for e in l], "r-")
  plt.plot([e[0] for e in l],
           [e[1] for e in l], "b.")
  plt.xlim([0, 9000])
  plt.ylim([0, 1500])
  plt.xlabel("distancia")
  plt.ylabel("coste")
  plt.show()



def pmx(parent_a, parent_b):
  """
  Method that giving two chromosomes returns a Partially-Mapped Crossover
  child

  @param parent_a    a list that represents a chromosome
  @param parent_b    a list that represents a chromosome
  @return mapped
  """
  assert(len(parent_a) == len(parent_b))
  n = len(parent_a)
  cross_points = sorted([random.randint(0, n) for _ in range(2)])
  swapped = _swap(parent_a, parent_b, cross_points)
  print(swapped)
  mapped = _map(swapped, cross_points)

  return mapped



def roulette(matrix, population, function_criteria="MIN"):
  """
  Returns a percentage piece to choice some chromosome from population.

  @param matrix               distance matrix
  @param population           list of chromosomes
  @param function_criteria    minimize/maximize fitness function
  @return res                 dict with probabilites to get some chromosome 
                              from population
  """
  chromosomes = []
  fit = fitness_all(matrix, population, function_criteria)
  total = 0
  # malformed solutions are not choosen to be selected
  for chromosome in population:
    if fit[tuple(chromosome)] == numpy.inf or \
       fit[tuple(chromosome)] == -numpy.inf:
      pass
    else:
      chromosomes.append(tuple(chromosome))
      total += fit[tuple(chromosome)]

  c = Counter(chromosomes)
  print("OCCURRENCES CHROMOSOMES: ")
  for e in c:
    print("{} -> {}".format(list(e), c[e]))
  print()
  res = {}
  i = 0
  for chromosome in set(chromosomes):
    if i == 0:
      aux = Fraction(fit[tuple(chromosome)] * \
                     c[tuple(chromosome)], total) * Fraction(360, 1)
      res[tuple(chromosome)] = (0, aux)
    else:
      aux_1 = aux + (Fraction(fit[tuple(chromosome)] * \
                              c[tuple(chromosome)], total) * \
                     Fraction(360, 1))
      res[tuple(chromosome)] = (aux, aux_1)
      aux = aux_1
    i = i + 1

  # https://matplotlib.org/examples/pie_and_polar_charts/index.html
  #labels = []
  #sizes = []
  #for chromosome in res:
  #  labels.append(chromosome)
  #  sizes.append((res[chromosome][1] - res[chromosome][0]) * 100)
  #fig1, ax1 = plt.subplots()
  #ax1.pie(sizes, labels=labels, autopct='%1.1f%%', startangle=90)
      
  # equal aspect ratio ensures that pie is drawn as a circle
  #ax1.axis("equal")
  #plt.show()
  #input()

  print("ROULETTE: ")
  for e in sorted(res.items(), key=lambda t: t[1]):
    print("{} -> {}".format(list(e[0]), e[1]))
  print()
  return res



def test_nsgaii_selection():
  chromosomes = ((5, 2), (1, 3), (6, 1), (3, 3), (2, 2))
  """
  chromosomes = set()
  while len(chromosomes) < 20:
    chromosomes.add((random.randint(0, 10),random.randint(0, 10)))
  """
  print(non_dominated_sort(chromosomes))
  """
  for front in non_dominated_sort(chromosomes):
    print(crowding_distance(front))
  """



def test_pareto_plot():
  #example_points = set([])
  #while len(example_points) < 30:
    #example_points.add((random.randint(0, 10)))
    #example_points.add((random.randint(0, 10), random.randint(0, 10)))
    #example_points.add((random.random() * 10, random.random() * 10))
    #example_points.add((random.randint(0, 10), random.randint(0, 10), random.randint(0, 10)))
  #example_points = list(example_points)
  #example_points = [(5, 9, 6), (3, 4, 7), (7, 10, 3), (0, 1, 4), \
  #                  (1, 6, 7), (8, 7, 1), (1, 8, 2), (8, 8, 9),  \
  #                  (3, 7, 3), (2, 10, 8), (10, 5, 5), (7, 0, 6), \
  #                  (0, 3, 6), (1, 3, 0), (8, 2, 10)]
  example_points = [(5, 2), (1, 3), (6, 1), (3, 3), (2, 2)]
  x = 0
  while x < len(example_points):
    example_points[x] = numpy.array(example_points[x])
    x = x + 1
  FUNCTION_DOMAINS = ("MIN", "MIN")
  #FUNCTION_DOMAINS = ("MIN", "MIN", "MAX")
  example = pareto_optimal(example_points, FUNCTION_DOMAINS)
  example_pareto_points = example[0]
  example_other_points = example[1]
  #print(example_pareto_points)
  #print(example_other_points)
  plot_pareto(example_pareto_points, example_other_points, FUNCTION_DOMAINS)



def test_pmx(chromosome_1=[2, 5, 8, 0, 1, 7, 4, 3, 6], \
             chromosome_2=[3, 7, 1, 2, 6, 4, 5, 8, 0]):
  """
  Method to test pmx() function

  @param chromosome_1    a list
  @param chromosome_2    a list
  """
  # test this with cut points [2, 7]
  print(chromosome_1, chromosome_2)
  offspring = pmx(chromosome_1, chromosome_2)
  print(offspring)



def test_pmx_deap(chromosome_1=[2, 5, 8, 0, 1, 7, 4, 3, 6], \
                  chromosome_2=[3, 7, 1, 2, 6, 4, 5, 8, 0]):
  """
  Method to test pmx() function in deap.tools

  @param chromosome_1    a list
  @param chromosome_2    a list
  """
  # test this with cut points [2, 7]
  t = tools.cxPartialyMatched(chromosome_1, chromosome_2)
  print(t)



def test_utils_ga_mono_functions():
  """
  Methods for test utils functions for GeneticAlgorithm
  """
  #print(300000000 > numpy.inf)
  #t = numpy.array([[1, 2, 3, 4], [5, 6, 7, 8], [1, 2, 3, 4], [5, 6, 7, 8]])
  #test = get_n_permutations(4, 24)
  #print(test)
  #print(t)
  #print(fitness(t, [(0, 3, 1, 2), (3, 2, 0, 1)]))
  #print(roulette(matrix=t, population=[(0, 3, 1, 2), (3, 2, 0, 1)]))
  #t = get_n_permutations(3, 100, repeated=True)
  #print(t)
  pass




if __name__ == "__main__":
  plot_jMetal_data()
  #test_nsgaii_selection()
  #test_pareto_plot()
  #test_pmx_deap()
  #test_pmx()
  pass
